/*
 *   This file is part of the Maximum library.
 *
 *   Maximum is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Maximum is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Maximum.  If not, see <http://www.gnu.org/licenses/>.
 */
using System;
using System.Diagnostics;

namespace Maximum
{
	/// <summary>
	/// Date with no time
	/// </summary>
	public class Date
	{
        private int m_year;
        private int m_month;
        private int m_day;
        
        /// <summary>
		/// Initialize to current date
		/// </summary>
		public Date()
		{
			DateTime now = DateTime.Now;
			Init( now.Year, now.Month, now.Day );
		}

		/// <summary>
		/// Initialize
		/// </summary>
		/// <param name="dt"></param>
		public Date( DateTime dt )
		{
			Init( dt.Year, dt.Month, dt.Day );
		}

		/// <summary>
		/// Create a new date
		/// </summary>
		/// <param name="year"></param>
		/// <param name="month"></param>
		/// <param name="day"></param>
		/// <exception cref="ArgumentOutOfRangeException"></exception>
		public Date( int year, int month, int day )
		{
			Init( year, month, day );
		}

		/// <summary>
		/// initialize a date from 20040101 format
		/// </summary>
		/// <param name="revdate"></param>
		public Date(int revdate)
		{
			m_year = revdate / 10000;
			m_month = (revdate - (m_year * 10000)) / 100;
			m_day = ((revdate - (m_year * 10000)) - (m_month * 100));
		}

		/// <summary>
		/// year part
		/// </summary>
		public int Year
		{
			get
			{
				return m_year;
			}
		}

		/// <summary>
		/// month part
		/// </summary>
		public int Month
		{
			get
			{
				return m_month;
			}
		}

		/// <summary>
		/// day part
		/// </summary>
		public int Day
		{
			get
			{
				return m_day;
			}
		}

		/// <summary>
		/// convert to datetime
		/// </summary>
		public DateTime AsDateTime
		{
			get
			{
				return new DateTime( m_year, m_month, m_day );
			}
		}

		public DateTime ToDateTime()
		{
			return AsDateTime;
		}

		public int ToRevInt()
		{
			return m_year * 10000 + m_month * 100 + m_day;
		}

		/// <summary>
		/// Extremely limited format cap
		/// </summary>
		/// <param name="frmt">"MMDDYY"</param>
		/// <returns></returns>
		public string Format( string frmt )
		{
			if ( frmt == "MMDDYY" )
			{
				Debug.Assert(1.ToString("00") == "01", "ToString format assumption failure");
				return m_month.ToString("00") + m_day.ToString("00") + TwoDigitYear(m_year).ToString("00");
			}
			throw new ArgumentException( "Can't format to " + frmt );
		}

		/// <summary>
		/// Return a four digit year from a two digit one
		/// </summary>
		/// <param name="year"></param>
		/// <returns></returns>
		public static int TwoDigitYear( int year )
		{
			Debug.Assert(Int32.Parse(year.ToString().Substring(2, 2)) == year % 100, "Conversion assumption failure");
			return year % 100;
		}

		public static bool operator >(Date d1, Date d2)
		{
			if (d1.m_year > d2.Year)
			{
				return true;
			}
			if (d1.m_year == d2.Year && (d1.m_month > d2.Month || (d1.m_month == d2.Month && d1.Day > d2.Day)))
			{
				return true;
			}
			return false;
		}

		public static bool operator <(Date d1, Date d2)
		{
			if (d1.m_year < d2.Year)
			{
				return true;
			}
			if (d1.m_year == d2.Year && (d1.m_month < d2.Month || (d1.m_month == d2.Month && d1.Day < d2.Day)))
			{
				return true;
			}
			return false;
		}

		public static bool operator >(Date d1, DateTime d2)
		{
			if ( d1.m_year > d2.Year )
			{
				return true;
			}
			if ( d1.m_year == d2.Year && (d1.m_month > d2.Month || (d1.m_month == d2.Month && d1.Day > d2.Day) ) )
			{
				return true;
			}
			return false;
		}

		public static bool operator <(Date d1, DateTime d2)
		{
			if ( d1.m_year < d2.Year )
			{
				return true;
			}
			if ( d1.m_year == d2.Year && (d1.m_month < d2.Month || (d1.m_month == d2.Month && d1.Day < d2.Day) ) )
			{
				return true;
			}
			return false;
		}

		public static bool operator <=(Date d1, DateTime d2)
		{
			return d1 < d2 || d1 == d2;
		}

		public static bool operator >=(Date d1, DateTime d2)
		{
			return d1 > d2 || d1 == d2;
		}

		public static bool operator <=(Date d1, Date d2)
		{
			return d1 < d2 || d1 == d2;
		}

		public static bool operator >=(Date d1, Date d2)
		{
			return d1 > d2 || d1 == d2;
		}

		/// <summary>
		/// Are the two dates equal?
		/// </summary>
		/// <param name="d1"></param>
		/// <param name="d2"></param>
		/// <returns></returns>
		public static bool operator ==( Date d1, Date d2 )
		{
			if ( (object)d1 == (object)d2 )
			{
				return true;
			}
			if ( (object)d1 == null || (object)d2 == null )
			{
				return false;
			}
			return d1.m_day == d2.Day && d1.m_month == d2.Month && d1.m_year == d2.Year;
		}

		/// <summary>
		/// Are the two dates equal?
		/// </summary>
		/// <param name="d1"></param>
		/// <param name="d2"></param>
		/// <returns></returns>
		public static bool operator ==( Date d1, DateTime d2 )
		{
			if ( (object)d1 == null )
			{
				return null == (object)d2;
			}
			return d1.Equals( d2 );
		}

		/// <summary>
		/// Are the two date different
		/// </summary>
		/// <param name="d1"></param>
		/// <param name="d2"></param>
		/// <returns></returns>
		public static bool operator !=(Date d1, Date d2)
		{
            if (null == (object)d1)
            {
                return null != (object)d2;
            }
			return ! d1.Equals( d2 );
		}

		/// <summary>
		/// Are the two date different
		/// </summary>
		/// <param name="d1"></param>
		/// <param name="d2"></param>
		/// <returns></returns>
		public static bool operator !=(Date d1, DateTime d2)
		{
            if (null == (object)d1)
            {
                return null != (object)d2;
            }
            return d1.m_day != d2.Day || d1.m_month != d2.Month || d1.m_year != d2.Year;
		}
		
		/// <summary>
		/// same as ==
		/// </summary>
		/// <param name="o"></param>
		/// <returns></returns>
		public override bool Equals(object o)
		{
			Date d2 = o as Date;
			if ( d2 == null )
			{
				return false;
			}
			return m_day == d2.m_day && m_month == d2.m_month && m_year == d2.m_year;
		}

		/// <summary>
		/// required override
		/// </summary>
		/// <returns></returns>
		public override int GetHashCode()
		{
			return base.GetHashCode();
		}

		public int AsRevInt
		{
			get	{ return m_year * 10000 + m_month * 100 + m_day; }
		}

		public override string ToString()
		{
			return m_month.ToString("00") + "/" + m_day.ToString("00") + "/" + m_year.ToString();
		}

		public static Date Parse(string dt)
		{
			int mo, dy, yr;

			if (dt.IndexOf(',') > -1)
			{
				switch (dt.Substring(0, 3))
				{
					case "Jan":
						mo = 1;
						break;
					case "Feb":
						mo = 2;
						break;
					case "Mar":
						mo = 3;
						break;
					case "Apr":
						mo = 4;
						break;
					case "May":
						mo = 5;
						break;
					case "Jun":
						mo = 6;
						break;
					case "Jul":
						mo = 7;
						break;
					case "Aug":
						mo = 8;
						break;
					case "Sep":
						mo = 9;
						break;
					case "Oct":
						mo = 10;
						break;
					case "Nov":
						mo = 11;
						break;
					case "Dec":
						mo = 12;
						break;
					default:
						throw new FormatException();
				}
				int pos = dt.IndexOf(' ') + 1;
				int cmaidx = dt.IndexOf(',');
				dy = Int32.Parse( StringHelper.MidStr(dt, pos, cmaidx));
				yr = Int32.Parse(dt.Substring(cmaidx + 1));

				return new Date(yr, mo, dy);
			}
			if (CharCount(dt, '/') != 2)
			{
				throw new FormatException();
			}
			string[] parts = dt.Split(new char[] { '/' });
			if (!Int32.TryParse(parts[0], out mo))
			{
				throw new FormatException();
			}
			if (!Int32.TryParse(parts[1], out dy))
			{
				throw new FormatException();
			}
			if (!Int32.TryParse(parts[2], out yr))
			{
				throw new FormatException();
			}
			return new Date(yr, mo, dy);
		}

		public static bool IsDate(string dt)
		{
			return Parse(dt) != null;
		}

		public static Date Now
		{
			get { return new Date(); }
		}
		
		private static int CharCount(string str, char ch)
		{
			int count = 0;
			for (int x = 0; x < str.Length; x++)
			{
				if (str[x] == ch)
				{
					count++;
				}
			}
			return count;
		}

		/// <summary>
		/// Init
		/// </summary>
		/// <param name="year"></param>
		/// <param name="month"></param>
		/// <param name="day"></param>
		private void Init(int year, int month, int day)
		{
			Debug.Assert( year > 1900, "Invalid year of " + year );
			Debug.Assert( month > 0 && month < 13, "Invalid month of " + month );
			Debug.Assert( day > 0 && day < 32, "Invalid day of " + day );

			if ( year < 1900 || month <=0 || month > 12 || day <= 0 || day > 31 )
			{
				throw new ArgumentOutOfRangeException("Date", "Invalid date (" + month + "/" + day + "/" + year + ")");
			}
			m_year = year;
			m_month = month;
			m_day = day;
		}

		public static bool IsDate(int year, int mo, int day)
		{
			return mo > 0 && mo < 12 && day > 0 && day < 32;
		}

		//public static bool RegressToWorkday(ref DateTime dtm)
		//{
		//    bool changed = false;

		//    while (IsHoliday(dtm) || IsWeekend(dtm))
		//    {
		//        dtm = dtm.AddDays(-1);
		//        changed = true;
		//    }
		//    return changed;
		//}

		//public static bool IsWeekend(DateTime dtm)
		//{
		//    if (dtm.DayOfWeek == DayOfWeek.Sunday || dtm.DayOfWeek == DayOfWeek.Saturday)
		//        return true;
		//    return false;
		//}

		//public bool IsHoliday()
		//{
		//    switch (m_month)
		//    {
		//        case 1:
		//            //	New Years
		//            //	RULE
		//            //	January 1.  If on a Saturday, the previous friday is off. If on a Sunday then Monday is off
		//            if (intDay == 1 && (DayOfWeek != DayOfWeek.Saturday && DayOfWeek != DayOfWeek.Sunday))
		//            {
		//                return true;
		//            }
		//            else if (intDay == 2 && dtm.DayOfWeek == DayOfWeek.Monday)
		//                return true;

		//            //	MLK Day
		//            //	RULE
		//            //	celebrated as third Monday in January, King's birthday is Jan 18.
		//            if (dtm.DayOfWeek == DayOfWeek.Monday)
		//            {
		//                DateTime mlkDay = new DateTime(dtm.Year, 1, 1);
		//                while (mlkDay.DayOfWeek != DayOfWeek.Monday)
		//                    mlkDay = mlkDay.AddDays(1);

		//                mlkDay = mlkDay.AddDays(14);

		//                if (dtm.Date == mlkDay.Date)
		//                    return true;
		//            }
		//            break;
		//        case 2:
		//            //	Presidents' Day
		//            //	RULE
		//            //	Third Monday in February.  Replaced Lincoln and Washington's birthday in 1971.
		//            if (dtm.DayOfWeek == DayOfWeek.Monday)
		//            {
		//                DateTime presDay = new DateTime(dtm.Year, 2, 1);
		//                while (presDay.DayOfWeek != DayOfWeek.Monday)
		//                    presDay = presDay.AddDays(1);

		//                presDay = presDay.AddDays(14);

		//                if (dtm.Date == presDay.Date)
		//                    return true;
		//            }
		//            break;
		//        case 5:
		//            //	Memorial Day
		//            //	RULE
		//            //	Last Monday in May. Originally May 31
		//            if (dtm.DayOfWeek == DayOfWeek.Monday)
		//            {
		//                if (dtm.AddDays(7).Month == 6)
		//                    return true;
		//            }
		//            break;
		//        case 7:
		//            // July 4th
		//            // RULE
		//            //	If on a Saturday, then friday is off
		//            //	If on a Sunday, then monday is off
		//            if (intDay == 4 && (dtm.DayOfWeek != DayOfWeek.Saturday && dtm.DayOfWeek != DayOfWeek.Sunday))
		//            {
		//                return true;
		//            }
		//            if (intDay == 3 && dtm.DayOfWeek == DayOfWeek.Friday)
		//            {
		//                return true;
		//            }
		//            if (intDay == 5 && dtm.DayOfWeek == DayOfWeek.Monday)
		//            {
		//                return true;
		//            }
		//            break;
		//        case 9:
		//            //	Labor Day
		//            //	RULE
		//            //	First Monday in September. In 1882..1883 it was celebrated on September 5.
		//            if (dtm.DayOfWeek == DayOfWeek.Monday)
		//            {
		//                if (dtm.AddDays(-7).Month == 8)
		//                    return true;
		//            }
		//            break;
		//        case 10:
		//            //	Columbus Day
		//            //	RULE
		//            //	1905 .. 1970 -> October 12
		//            //	1971 .. now  -> Second Monday in October.
		//            if (dtm.DayOfWeek == DayOfWeek.Monday)
		//            {
		//                DateTime columbusDay = new DateTime(dtm.Year, 10, 1);
		//                while (columbusDay.DayOfWeek != DayOfWeek.Monday)
		//                    columbusDay = columbusDay.AddDays(1);

		//                columbusDay = columbusDay.AddDays(7);

		//                if (dtm.Date == columbusDay.Date)
		//                    return true;
		//            }
		//            break;
		//        case 11:
		//            // Veterans' Day
		//            // RULE
		//            // November 11.  If on a Saturday, the previous friday is off. If on a Sunday then Monday is off
		//            if (intDay == 11 && dtm.DayOfWeek != DayOfWeek.Saturday && dtm.DayOfWeek != DayOfWeek.Sunday)
		//            {
		//                return true;
		//            }
		//            if (intDay == 10 && dtm.DayOfWeek == DayOfWeek.Friday)
		//            {
		//                return true;
		//            }
		//            if (intDay == 12 && dtm.DayOfWeek == DayOfWeek.Monday)
		//            {
		//                return true;
		//            }

		//            //  Thanksgiving 
		//            //	RULE
		//            //	1621         -> first Thanksgiving, precise date unknown.
		//            //	1622         -> was no Thanksgiving.
		//            //	1623 .. 1675 -> precise date unknown.
		//            //	1676 .. 1862 -> June 29.
		//            //	1863 .. 1938 -> last Thursday of November.
		//            //	1939 .. 1941 -> 2nd to last Thursday of November.
		//            //	1942 .. now  -> 4th Thursday of November.
		//            if (dtm.DayOfWeek == DayOfWeek.Thursday)
		//            {
		//                DateTime thanksgivingDay = new DateTime(dtm.Year, 11, 1);
		//                while (thanksgivingDay.DayOfWeek != DayOfWeek.Thursday)
		//                    thanksgivingDay = thanksgivingDay.AddDays(1);

		//                thanksgivingDay = thanksgivingDay.AddDays(21);

		//                if (dtm.Date == thanksgivingDay.Date)
		//                    return true;
		//            }
		//            else if (dtm.DayOfWeek == DayOfWeek.Friday)
		//            {
		//                // the day after is also a bank holiday
		//                DateTime dayAfterTg = new DateTime(dtm.Year, 11, 1);
		//                while (dayAfterTg.DayOfWeek != DayOfWeek.Thursday)
		//                    dayAfterTg = dayAfterTg.AddDays(1);

		//                dayAfterTg = dayAfterTg.AddDays(22);

		//                if (dtm.Date == dayAfterTg.Date)
		//                    return true;
		//            }
		//            break;
		//        case 12:
		//            //	Christmas
		//            //	RULE
		//            //	 If on a saturday, then friday is off
		//            //	 If on a sunday, then monday is off
		//            if (intDay == 25 && (dtm.DayOfWeek != DayOfWeek.Saturday && dtm.DayOfWeek != DayOfWeek.Sunday))
		//            {
		//                return true;
		//            }
		//            else if (intDay == 24 && dtm.DayOfWeek == DayOfWeek.Friday)
		//            {
		//                // the 25th is on saturday
		//                return true;
		//            }
		//            else if (intDay == 26 && dtm.DayOfWeek == DayOfWeek.Monday)
		//            {
		//                // the 25th is on sunday
		//                return true;
		//            }

		//            //	New Years
		//            //	RULE
		//            //	January 1.  If on a Saturday, the previous friday is off
		//            if (dtm.Day == 31 && dtm.DayOfWeek == DayOfWeek.Friday)
		//            {
		//                // the friday of Dec 31 is a holiday (jan 1 is on saturday)
		//                return true;
		//            }
		//            break;
		//    }
		//    return false;
		//}	
	}
}
