/*
 *   This file is part of the Maximum library.
 *
 *   Maximum is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Maximum is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Maximum.  If not, see <http://www.gnu.org/licenses/>.
 */
using System;
using System.IO;
using System.Collections.Specialized;
using System.Collections.ObjectModel;
using System.Collections;
using System.Configuration;
using System.Text;
using System.Diagnostics;

namespace Maximum
{
	/// <summary>
	/// Error logging class
	/// </summary>
	public class LogFile
	{
		private static LogFile m_syslog = new LogFile(ConfigurationManager.AppSettings["LOG_FILE"]);

		private string m_filename;
		private object myLock = new object();
		private int m_messageCount;

		public int MessageCount
		{
			get { return m_messageCount; }
		}

		public LogFile(string filename)
		{
			if (null == filename || "" == filename)
			{
				filename = "LOGFILE.txt";
			}
			m_filename = filename;
		}

		private void Write( string msg )
		{
			lock( myLock )
			{
				m_messageCount++;
				StreamWriter writer = null;
				try
				{
					writer = File.AppendText( m_filename );
					writer.WriteLine(msg);
				}
				catch(Exception)
				{
				}
				finally
				{
					if ( null != writer )
					{
						writer.Close();
					}
				}
			}
			try
			{
				//ELF.SmtpMailer mailer = new ELF.SmtpMailer();
				//mailer.Send( "LBTS", "johnga@dor.wa.gov", "johnga@dor.wa.gov", "LBTS ERROR", msg );
			}
			catch( Exception )
			{
				// dont' do anything
			}
		}

		public void WriteLog(string caller, Exception ex)
		{
			WriteLog(caller, null, ex);
		}

		public void WriteLog( string user, string caller, Exception ex )
		{
			StringBuilder sb = new StringBuilder();

			sb.Append( "EXCEPTION " );
			sb.Append( DateTime.Now.ToString() );
			if (null != user)
			{
				sb.Append("\t" + caller + " User ID: " + user + "\r\n");
			}
			sb.Append("\r\n\t");
			sb.Append(ex.Message);
			sb.Append("\r\n\r\n\t");
			sb.Append(ex.StackTrace.Replace("\r\n", "\r\n\t"));
			sb.Append("\r\n---------------------------------------------\r\n\r\n");
			Write(sb.ToString());
		}		

		public void WriteLog( string user, string caller, string msg )
		{
			StringBuilder sb = new StringBuilder();
			sb.Append("INFO ");
			sb.Append(DateTime.Now.ToString());
			sb.Append("\r\n\t" + caller + " User ID: " + ((user == null)? "unknown" : user) );
			sb.Append("\r\n\t");
			sb.Append(msg);
			sb.Append("\r\n---------------------------------------------\r\n\r\n");
			Write( sb.ToString() );
		}		

		public void WriteLog( string caller, string message )
		{
			StringBuilder sb = new StringBuilder();
			sb.Append("INFO ");
			sb.Append(DateTime.Now.ToString() + "\r\n");
			sb.Append(caller + ": " + message);
			sb.Append("\r\n---------------------------------------------\r\n\r\n");
			Write( sb.ToString() );
		}

		public void Assert( bool cond, string msg )
		{
			if ( ! cond )
			{
				StringBuilder sb = new StringBuilder();
				sb.Append("ASSERT");
				sb.Append(DateTime.Now.ToString()+"\r\n");
				StackTrace stk = new StackTrace();
				sb.Append(msg+"\r\n");
				sb.Append(stk.ToString().Replace("\r\n", "\r\n\t"));
				sb.Append("\r\n---------------------------------------------\r\n\r\n");
				Write( sb.ToString() );
			}
		}

		public string DisplayError(string error, string explanation)
		{
			StringBuilder sb = new StringBuilder();
			sb.Append("<br><TABLE id=\"ErrorTable\" borderColor=\"dimgray\" cellSpacing=\"0\" cellPadding=\"3\" width=\"475\" align=\"center\" bgColor=\"gainsboro\" border=\"1\">");
			sb.Append("<TBODY>");
			sb.Append("<TR>");
			sb.Append("<TD bgColor=\"silver\"><FONT color=\"red\"><B>");
			sb.Append(error);
			sb.Append("</B></FONT></TD>");
			sb.Append("</TR>");
			sb.Append("<TR>");
			sb.Append("<TD><STRONG><FONT size=\"2\">");
			sb.Append(explanation);
			sb.Append("</FONT></STRONG></TD>");
			sb.Append("</TR></TBODY></TABLE><BR>");
				
			return sb.ToString();
		}

		public static void SysWriteLog(string caller, Exception ex)
		{
			m_syslog.WriteLog(caller, ex);
		}

		public static void SysWriteLog(string user, string caller, Exception ex)
		{
			m_syslog.WriteLog(user, caller, ex);
		}

		public static void SysWriteLog(string user, string caller, string msg)
		{
			m_syslog.WriteLog(user, caller, msg);
		}

		public static void SysWriteLog(string caller, string message)
		{
			m_syslog.WriteLog(caller, message);
		}

		public static void SysAssert(bool cond, string msg)
		{
			m_syslog.Assert(cond, msg);
		}

		public static string SysDisplayError(string error, string explanation)
		{
			return m_syslog.DisplayError(error, explanation);
		}
	}
}
