/*
 *   This file is part of the Maximum library.
 *
 *   Maximum is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Maximum is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Maximum.  If not, see <http://www.gnu.org/licenses/>.
 */
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;

namespace Maximum.NeuralNet.Backprop
{
	public class RawData
	{
		private List<double> m_data = new List<double>();
		private DataStats m_stats;

		public RawData()
		{
		}

		public void Append(double i)
		{
			m_data.Add(i);
		}

		public DataStats Stats
		{
			get
			{
				if (null == m_stats)
				{
					m_stats = new DataStats(m_data);
				}
				return m_stats;
			}
		}

		public int Count
		{
			get { return m_data.Count; }
		}

		public double this [int idx]
		{
			get { return m_data[idx]; }
		}

		public double[] ToArray()
		{
			return m_data.ToArray();
		}
	}

	public class NormalizedData
	{
		private DataStats m_stats;
		private List<double> m_data = new List<double>();
		private bool m_useSquashing;

		private void Add(double d)
		{
			Debug.Assert(!Double.IsInfinity(d));
			Debug.Assert(!Double.IsNaN(d));

			double min = m_stats.Min;
			if (m_useSquashing)
			{
				m_data.Add(1.0 / (1.0 + System.Math.Exp(-(d - m_stats.Mean) / m_stats.StDevSample)));
			}
			else
			{
				m_data.Add((d - min) / (m_stats.Max - min));
			}
			Debug.Assert(m_data[m_data.Count - 1] >= 0);
		}

		public NormalizedData(RawData raw, bool squash)
		{
			m_stats = raw.Stats;
			m_useSquashing = squash;
			Debug.Assert(m_stats.StDevSample != 0);
			int n = raw.Count;

			for (int x = 0; x < n; x++)
			{
				Add(raw[x]);
			}
		}

		public NormalizedData(double[] raw, bool squash)
		{
			m_stats = new DataStats(raw);
			m_useSquashing = squash;
			Debug.Assert(m_stats.StDevSample != 0);
			int n = raw.Length;
			for (int x = 0; x < n; x++)
			{
				Debug.Assert(!Double.IsNaN(raw[x]));
				Add(raw[x]);
			}
		}

		public int Count
		{
			get { return m_data.Count; }
		}

		public double this[int idx]
		{
			get { return m_data[idx]; }
		}

		public double[] ToArray()
		{
			return m_data.ToArray();
		}

		public double DeNormalize(double y)
		{
			if (m_useSquashing)
			{
				//-std * ln( 1.0/y - 1 ) + mean = x
				return -m_stats.StDevSample * Math.Log(1.0 / y - 1) + m_stats.Mean;
				//return 1.0 / (1.0 + Math.Exp(- (x - m_rd.MeanHi) / m_rd.StdHi));
			}
			else
			{
				return y * (m_stats.Max - m_stats.Min) + m_stats.Min;
			}
		}
	}
}
