/*
 *   This file is part of the Maximum library.
 *
 *   Maximum is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Maximum is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Maximum.  If not, see <http://www.gnu.org/licenses/>.
 */
using System;
using System.Data;
//using System.Data.SqlClient;
using System.Diagnostics;

using MySql.Data;
using MySql.Data.MySqlClient;

namespace Maximum.NeuralNet.Backprop
{
	public class NeuralNetworkException : Exception
	{
		public NeuralNetworkException( string msg ) : base( msg )
		{
		}
	}

	enum Layer
	{
		INPUT = 0,
		L1 = 1,
		L2 = 2,
		OUTPUT = 3
	}

	/// <summary>
	/// 
	/// </summary>
	public class NeuralNetwork
	{
        private int m_networkId = -1;
        private double m_learnRate = .2;
        private double m_momentum = .05;
        private double m_wtRange = .2;
        private bool m_useAdaptiveLR;
        private bool m_useAdaptiveMom;
        private double m_previousError = 0.0;

        private double[] m_input; // nodes
        private double[] m_layer1;
        private double[] m_layer2;
        private double[] m_output;

        private double[] m_errors1; // error for each node
        private double[] m_errors2;
        private double[] m_errorsOut;

        private double[] m_wt1ToIn; // weights from layer 1 to inputs
        private double[] m_delta1ToIn; // deltas from layer 1 to inputs

        private double[] m_wt2To1; // weights from layer 2 to layer 1
        private double[] m_delta2To1; // deltas from layer 2 to layer 1

        private double[] m_wtOutTo2; // weights from output to layer 3
        private double[] m_deltaOutTo2; // deltas from output to layer 3

        private static Random m_rand = new Random();

		public int NumConnections
		{
			get { return m_wt1ToIn.Length + m_wt2To1.Length + m_wtOutTo2.Length; }
		}

        private void BuildNetwork(int input, int h1, int h2, int output)
		{
			// all layers must be populated
			if (input == 0 || h1 == 0 || h2 == 0 || output == 0)
			{
				throw new NeuralNetworkException("Layers with zero nodes not allowed");
			}
			m_input = new double[input];
			m_layer1 = new double[h1];
			m_layer2 = new double[h2];
			m_output = new double[output];
			
			m_errors1 = new double[h1];
			m_errors2 = new double[h2];
			m_errorsOut = new double[output];
			
			m_wt1ToIn = new double[input * h1]; // weights from layer 1 to inputs
			m_delta1ToIn = new double[input * h1]; // deltas from layer 1 to inputs
			
			m_wt2To1 = new double[h1 * h2]; // weights from layer 2 to layer 1
			m_delta2To1 = new double[h1 * h2]; // deltas from layer 2 to layer 1
			
			m_wtOutTo2 = new double[output * h2]; // weights from output to layer 3
			m_deltaOutTo2 = new double[output * h2]; // deltas from output to layer 3
			
			RandomizeWeights();
		}

		public NeuralNetwork(int input, int h1, int h2, int output, bool adaptiveLearnRate, bool adaptiveMomentum)
		{
			m_useAdaptiveLR = adaptiveLearnRate;
			m_useAdaptiveMom = adaptiveMomentum;
			BuildNetwork( input, h1, h2, output );
		}

		public NeuralNetwork( int networkId, MySqlConnection con )
		{
			MySqlCommand cmd = new MySqlCommand( "selNetwork", con );
			cmd.CommandType = CommandType.StoredProcedure;
			//cmd.Parameters.Add( "RETURN_VALUE", SqlDbType.Int );
			//cmd.Parameters["RETURN_VALUE"].Direction = ParameterDirection.ReturnValue;
			cmd.Parameters.AddWithValue( "_NETWORK_ID", networkId );

			MySqlDataReader reader = cmd.ExecuteReader();
			if ( ! reader.Read() )
			{
				reader.Close();
				throw new NeuralNetworkException("Network " + networkId + " not found");
			}
			m_networkId = networkId;
			m_learnRate = (double)reader["LEARN_RATE_NUM"];
			m_momentum = (double)reader["MOMENTUM_NUM"];
			m_wtRange = (double)reader["WEIGHT_RANGE_NUM"];
			m_useAdaptiveLR = (UInt64)reader["ADAPT_LEARN_RATE_IND"] != 0;
			m_useAdaptiveMom = (UInt64)reader["ADAPT_MOM_IND"] != 0;
			m_previousError = (double)reader["PREV_ERROR_NUM"];
			int input = (int)reader["IN_NODE_NUM"];
			int h1 = (int)reader["L1_NODE_NUM"];
			int h2 = (int)reader["L2_NODE_NUM"];
			int output = (int)reader["OUT_NODE_NUM"];
			reader.Close();

			BuildNetwork( input, h1, h2, output );

			cmd = new MySqlCommand( "selNodes", con );
			cmd.CommandType = CommandType.StoredProcedure;
			//cmd.Parameters.Add( "RETURN_VALUE", SqlDbType.Int );
			//cmd.Parameters["RETURN_VALUE"].Direction = ParameterDirection.ReturnValue;
			cmd.Parameters.AddWithValue( "_NETWORK_ID", networkId );

			reader = cmd.ExecuteReader();
			if ( ! reader.Read() )
			{
				// just use the newly created network;
				reader.Close();
				return;
			}
			do
			{
				int idx = (int)reader["INDEX_NUM"];
				double val = (double)reader["VAL_NUM"];

				switch ( (Layer)(sbyte)reader["LAYER_NUM"] )
				{
					case Layer.INPUT:
						m_input[idx] = val;
						break;
					case Layer.L1:
						m_layer1[idx] = val;
						break;
					case Layer.L2:
						m_layer2[idx] = val;
						break;
					case Layer.OUTPUT:
						m_output[idx] = val;
						break;
					default:
						throw new NeuralNetworkException("Unexpected layer num in nodes of " + reader["LAYER_NUM"].ToString() );
				}
			}
			while( reader.Read() );
			reader.Close();

			cmd = new MySqlCommand( "selConns", con );
			cmd.CommandType = CommandType.StoredProcedure;
			//cmd.Parameters.Add( "RETURN_VALUE", SqlDbType.Int );
			//cmd.Parameters["RETURN_VALUE"].Direction = ParameterDirection.ReturnValue;
            cmd.Parameters.AddWithValue("_NETWORK_ID", networkId);

			reader = cmd.ExecuteReader();
			if ( ! reader.Read() )
			{
				reader.Close();
				throw new NeuralNetworkException("Connections for network " + networkId + " not found");
			}
			do
			{
				int idx = (int)reader["INDEX_NUM"];
				double weight = (double)reader["WEIGHT_NUM"];
				double delta = (double)reader["DELTA_NUM"];

				switch ( (Layer)(sbyte)reader["LAYER_NUM"] )
				{
					case Layer.L1:
						this.m_wt1ToIn[idx] = weight;
						this.m_delta1ToIn[idx] = delta;
						break;
					case Layer.L2:
						this.m_wt2To1[idx] = weight;
						this.m_delta2To1[idx] = delta;
						break;
					case Layer.OUTPUT:
						this.m_wtOutTo2[idx] = weight;
						this.m_deltaOutTo2[idx] = delta;
						break;
					default:
						throw new NeuralNetworkException("Unexpected layer num in nodes of " + reader["LAYER_NUM"].ToString() );
				}
			}
			while( reader.Read() );
			reader.Close();			
		}

		public void Write( MySqlConnection con, MySqlTransaction trans )
		{
			bool update = true;
			MySqlCommand cmd = null;

			if ( m_networkId < 0 )
			{
				update = false;
				cmd = new MySqlCommand( "insNetwork", con, trans );
				cmd.CommandType = CommandType.StoredProcedure;
				//cmd.Parameters.Add("RETURN_VALUE", SqlDbType.Int);
				//cmd.Parameters["RETURN_VALUE"].Direction = ParameterDirection.ReturnValue;
                cmd.Parameters.Add("_NETWORK_ID", MySqlDbType.Int32);
				cmd.Parameters["_NETWORK_ID"].Direction = ParameterDirection.Output;
                cmd.Parameters.Add("_LEARN_RATE_NUM", MySqlDbType.Double).Value = m_learnRate;
                cmd.Parameters.Add("_MOMENTUM_NUM", MySqlDbType.Double).Value = m_momentum;
                cmd.Parameters.Add("_WEIGHT_RANGE_NUM", MySqlDbType.Double).Value = m_wtRange;
                cmd.Parameters.Add("_ADAPT_LEARN_RATE_IND", MySqlDbType.Bit).Value = m_useAdaptiveLR;
                cmd.Parameters.Add("_ADAPT_MOM_IND", MySqlDbType.Bit).Value = m_useAdaptiveMom;
                cmd.Parameters.Add("_PREV_ERROR_NUM", MySqlDbType.Double).Value = m_previousError;
                cmd.Parameters.AddWithValue("_IN_NODE_NUM", m_input.Length);
                cmd.Parameters.AddWithValue("_L1_NODE_NUM", m_layer1.Length);
                cmd.Parameters.AddWithValue("_L2_NODE_NUM", m_layer2.Length);
                cmd.Parameters.AddWithValue("_OUT_NODE_NUM", m_output.Length);
				cmd.ExecuteNonQuery();
				//if ( 0 != (int)cmd.Parameters["RETURN_VALUE"].Value )
				//{
				//	throw new Exception("insNetwork returned " + cmd.Parameters["RETURN_VALUE"].Value );
				//}
				m_networkId = (int)cmd.Parameters["_NETWORK_ID"].Value;
			}
			if ( update )
			{
				cmd = new MySqlCommand( "updNode", con, trans );
			}
			else
			{
				cmd = new MySqlCommand( "insNode", con, trans );
			}
			cmd.CommandType = CommandType.StoredProcedure;
			//cmd.Parameters.Add("RETURN_VALUE", SqlDbType.Int);
			//cmd.Parameters["RETURN_VALUE"].Direction = ParameterDirection.ReturnValue;
            cmd.Parameters.AddWithValue("_NETWORK_ID", m_networkId);
			cmd.Parameters.Add("_LAYER_NUM", MySqlDbType.Int16);
			cmd.Parameters.Add("_INDEX_NUM", MySqlDbType.Int32);
			cmd.Parameters.Add("_VAL_NUM", MySqlDbType.Double);

			cmd.Parameters["_LAYER_NUM"].Value = Layer.INPUT;		
			for (int x = 0; x < m_input.Length; x++)
			{
				cmd.Parameters["_INDEX_NUM"].Value = x;
				cmd.Parameters["_VAL_NUM"].Value = m_input[x];
				cmd.ExecuteNonQuery();
				//if ( 0 != (int)cmd.Parameters["RETURN_VALUE"].Value )
				//{
				//	throw new Exception(cmd.CommandText + " returned " + cmd.Parameters["RETURN_VALUE"].Value );
				//}
			}
			cmd.Parameters["_LAYER_NUM"].Value = Layer.L1;		
			for (int x = 0; x < m_layer1.Length; x++)
			{
				cmd.Parameters["_INDEX_NUM"].Value = x;
				cmd.Parameters["_VAL_NUM"].Value = m_layer1[x];
				cmd.ExecuteNonQuery();
				//if ( 0 != (int)cmd.Parameters["RETURN_VALUE"].Value )
				//{
				//	throw new Exception(cmd.CommandText + " returned " + cmd.Parameters["RETURN_VALUE"].Value );
				//}
			}
			cmd.Parameters["_LAYER_NUM"].Value = Layer.L2;		
			for (int x = 0; x < m_layer2.Length; x++)
			{
				cmd.Parameters["_INDEX_NUM"].Value = x;
				cmd.Parameters["_VAL_NUM"].Value = m_layer2[x];
				cmd.ExecuteNonQuery();
				//if ( 0 != (int)cmd.Parameters["RETURN_VALUE"].Value )
				//{
				//	throw new Exception(cmd.CommandText + " returned " + cmd.Parameters["RETURN_VALUE"].Value );
				//}
			}
			cmd.Parameters["_LAYER_NUM"].Value = Layer.OUTPUT;		
			for (int x = 0; x < m_output.Length; x++)
			{
				cmd.Parameters["_INDEX_NUM"].Value = x;
				cmd.Parameters["_VAL_NUM"].Value = m_output[x];
				cmd.ExecuteNonQuery();
				//if ( 0 != (int)cmd.Parameters["RETURN_VALUE"].Value )
				//{
				//	throw new Exception(cmd.CommandText + " returned " + cmd.Parameters["RETURN_VALUE"].Value );
				//}
			}

			if ( update )
			{
				cmd = new MySqlCommand( "updConn", con, trans );
			}
			else
			{
				cmd = new MySqlCommand( "insConn", con, trans );
			}
			cmd.CommandType = CommandType.StoredProcedure;
			//cmd.Parameters.Add("RETURN_VALUE", SqlDbType.Int);
			//cmd.Parameters["RETURN_VALUE"].Direction = ParameterDirection.ReturnValue;
            cmd.Parameters.AddWithValue("_NETWORK_ID", m_networkId);
			cmd.Parameters.Add("_LAYER_NUM", MySqlDbType.Int16);
			cmd.Parameters.Add("_INDEX_NUM", MySqlDbType.Int32);
			cmd.Parameters.Add("_WEIGHT_NUM", MySqlDbType.Double);
			cmd.Parameters.Add("_DELTA_NUM", MySqlDbType.Double);

			cmd.Parameters["_LAYER_NUM"].Value = Layer.L1;
			for (int x = 0; x < m_wt1ToIn.Length; x++)
			{
				cmd.Parameters["_INDEX_NUM"].Value = x;
				cmd.Parameters["_WEIGHT_NUM"].Value = m_wt1ToIn[x];
				cmd.Parameters["_DELTA_NUM"].Value = m_delta1ToIn[x];
				cmd.ExecuteNonQuery();
				//if ( 0 != (int)cmd.Parameters["RETURN_VALUE"].Value )
				//{
				//	throw new Exception(cmd.CommandText + " returned " + cmd.Parameters["RETURN_VALUE"].Value );
				//}
			}
			cmd.Parameters["_LAYER_NUM"].Value = Layer.L2;
			for (int x = 0; x < m_wt2To1.Length; x++)
			{
				cmd.Parameters["_INDEX_NUM"].Value = x;
				cmd.Parameters["_WEIGHT_NUM"].Value = m_wt2To1[x];
				cmd.Parameters["_DELTA_NUM"].Value = m_delta2To1[x];
				cmd.ExecuteNonQuery();
				//if ( 0 != (int)cmd.Parameters["RETURN_VALUE"].Value )
				//{
				//	throw new Exception(cmd.CommandText + " returned " + cmd.Parameters["RETURN_VALUE"].Value );
				//}
			}
			cmd.Parameters["_LAYER_NUM"].Value = Layer.OUTPUT;
			for (int x = 0; x < m_wtOutTo2.Length; x++)
			{
				cmd.Parameters["_INDEX_NUM"].Value = x;
				cmd.Parameters["_WEIGHT_NUM"].Value = m_wtOutTo2[x];
				cmd.Parameters["_DELTA_NUM"].Value = m_deltaOutTo2[x];
				cmd.ExecuteNonQuery();
				//if ( 0 != (int)cmd.Parameters["RETURN_VALUE"].Value )
				//{
				//	throw new Exception(cmd.CommandText + " returned " + cmd.Parameters["RETURN_VALUE"].Value );
				//}
			}
		}

		public void Write( MySqlConnection con )
		{
			MySqlTransaction trans = null;
			try
			{
				trans = con.BeginTransaction();
				Write( con, trans );
				trans.Commit();
			}
			catch ( Exception )
			{
				if ( null != trans )
				{
					trans.Rollback();
				}
				throw;
			}
		}

		public void Delete(MySqlConnection con)
		{
			if (m_networkId < 0)
			{
				return;
			}
			MySqlCommand cmd = new MySqlCommand("delNetwork", con);
			cmd.CommandType = CommandType.StoredProcedure;
			cmd.Parameters.AddWithValue("_NETWORK_ID", m_networkId);
			cmd.ExecuteNonQuery();
			m_networkId = -1;
		}

		public void Activate( double[] input, double[] output )
		{
			if ( input.Length != m_input.Length || output.Length != m_output.Length )
			{
				throw new NeuralNetworkException("input or output size mismatch");
			}
			Array.Copy(input, m_input, input.Length);
			Activate();
			for ( int x = 0; x < output.Length; x++ )
			{
				output[x] = m_output[x];
			}
		}

		/// <summary>  
		/// perform an activation of the network
		/// </summary>
		public void Activate()
		{
			ActivateLayer(m_input, m_layer1, m_wt1ToIn);
			ActivateLayer(m_layer1, m_layer2, m_wt2To1);
			ActivateLayer(m_layer2, m_output, m_wtOutTo2);
		}
		
		/// <summary>  
		/// layer1 is "behind" layer 2.  If layer1 is Input, then layer2
		/// is hidden layer 1.
		/// </summary>
		private void ActivateLayer(double[] layer1, double[] layer2, double[] wt)
		{
			double dSum;
			int iWtPos = 0;
			
			// for each node in layer 1
			for (int x = 0; x < layer2.Length; x++)
			{
				dSum = 0.0;
				// calculate the ouput for this node	
				// for all the nodes connected to this one
				for (int y = 0; y < layer1.Length; y++)
				{
					// sum the weights * outputs.  note that the weights are
					dSum += layer1[y] * wt[iWtPos++];
//					if ( wt[iWtPos-1].Equals( Double.NaN ) )
//					{
//						throw new NeuralNetworkException("NaN: Layer=" + layer1[y] + "; weight=" + wt[iWtPos-1]);
//					}
//					if ( layer1[y].Equals( Double.NaN ) )
//					{
//						throw new NeuralNetworkException("NaN: Layer=" + layer1[y] + "; weight=" + wt[iWtPos-1]);
//					}
//					if ( dSum.Equals( Double.NaN ) )
//					{
//						throw new NeuralNetworkException("NaN: Layer=" + layer1[y] + "; weight=" + wt[iWtPos-1]);
//					}
					Debug.Assert( ! wt[iWtPos-1].Equals( Double.NaN ) );
					Debug.Assert( ! layer1[y].Equals( Double.NaN ) );
					Debug.Assert( ! dSum.Equals(Double.NaN) );
				}
				// set the sum for the current node
				// [0,1]
				layer2[x] = 1.0 / (1 + Math.Exp(-dSum));
				// [0,1]
				//layer2[x] = 1.0 / (1 + Math.Exp(-2 * dSum));
				// [-1,1]
				//layer2[x] = Math.Tanh(dSum);

//				if ( layer2[x].Equals( Double.NaN ) )
//				{
//					throw new NeuralNetworkException("NaN: Sum layer[x]=" + layer2[x]);
//				}
				Debug.Assert( ! layer2[x].Equals( Double.NaN ) );
			}
		}

		public double CalcError( double[] input, double[] output )
		{
			double dErrTotal = 0;

			Array.Copy(input, m_input, input.Length);
			Activate();
						
			// calculate the final error
			for (int x = 0; x < m_output.Length; x++)
			{
				// calculate the absolute error for the output layer
				double err = output[x] - m_output[x];
				
				// save the squared error
				dErrTotal += err * err;
			}

			return dErrTotal;
		}

		/// <summary>  Perform a training iteration on one input/output set.  Training continues until
		/// the total error is below the specified threshhold or halt() is called.
		/// *
		/// </summary>
		/// <param name="in">The input values.  Must be at least as large as the input layer.
		/// </param>
		/// <param name="out">The output values.
		/// </param>
		/// <param name="dErrTarget">Training should continue until the error is below this value.
		/// </param>
		/// <param name="maxCycles">maxiumum number of training loops</param>
		/// <exception cref="">NeuralNetworkException Thrown if the weights explode or go to zero.
		/// 
		/// </exception>
		public double Train(double[] input, double[] output, double dErrTarget, int maxCycles)
		{
			int delay = 0;
			int iWtPos;
			double dErrTotal;

			for (int x = 0; x < input.Length; x++)
			{
				m_input[x] = input[x];
			}
			do 
			{
				// set the errors to 0
				SetArray(m_errors1, 0.0);
				SetArray(m_errors2, 0.0);
				
				// set the input
				if (input.Length != m_input.Length)
				{
					throw new NeuralNetworkException("Array size doesn't match number of input nodes");
				}				
				// activate
				Activate();
				
				iWtPos = 0;
				dErrTotal = 0.0;
				
				// calculate the final error
				for (int x = 0; x < m_output.Length; x++)
				{
					// calculate the absolute error for the output layer
					m_errorsOut[x] = output[x] - m_output[x];
					Debug.Assert( ! m_errorsOut[x].Equals( Double.NaN ) );
//					if ( m_errorsOut[x].Equals( Double.NaN ) )
//					{
//						throw new NeuralNetworkException("Errors NaN: output=" + output[x]);
//					}
					
					// save the squared error
					dErrTotal += m_errorsOut[x] * m_errorsOut[x];
					Debug.Assert( ! dErrTotal.Equals(Double.NaN) );
//					if ( dErrTotal.Equals(Double.NaN) )
//					{
//						throw new NeuralNetworkException("Total error NaN: error=" + m_errorsOut[x]);
//					}
					
					// back prop the error, update the weights
					for (int y = 0; y < m_layer2.Length; y++)
					{
						// back prop the error
						m_errors2[y] += m_errorsOut[x] * m_wtOutTo2[iWtPos];
						
						// update the delta and weight
						m_wtOutTo2[iWtPos] += (m_deltaOutTo2[iWtPos] = m_learnRate * m_errorsOut[x] * m_layer2[y] + m_momentum * m_deltaOutTo2[iWtPos]);
						//
						// need to check for exploding weights here
						//
						if (m_wtOutTo2[iWtPos] > 100.0 || m_wtOutTo2[iWtPos] < -100.0)
						{
							throw new NeuralNetworkException("Exploding weights");
						}
						if (m_wtOutTo2[iWtPos] == 0.0)
						{
							throw new NeuralNetworkException("Weight underflow");
						}
						iWtPos++;
					}
				}
				iWtPos = 0;
				// hidden layer 2
				for (int x = 0; x < m_layer2.Length; x++)
				{
					if (m_layer2[x] > 0)
					{
						m_errors2[x] *= m_layer2[x] * (1.0 - m_layer2[x]);
					}
					else
					{
						m_errors2[x] *= - m_layer2[x] * (1.0 + m_layer2[x]);
					}
					// back prop the error, update the weights
					for (int y = 0; y < m_layer1.Length; y++)
					{
						m_errors1[y] += m_errors2[x] * m_wt2To1[iWtPos];
						// update the delta and weight
						m_wt2To1[iWtPos] += (m_delta2To1[iWtPos] = m_learnRate * m_errors2[x] * m_layer1[y] + m_momentum * m_delta2To1[iWtPos]);
						//
						// need to check for exploding weights here
						//
						if (m_wt2To1[iWtPos] > 100.0 || m_wt2To1[iWtPos] < - 100.0)
						{
							throw new NeuralNetworkException("Exploding weights");
						}
						if (m_wt2To1[iWtPos] == 0.0)
						{
							throw new NeuralNetworkException("Weight underflow");
						}
						iWtPos++;
					}
				}
				iWtPos = 0;
				// hidden layer 1
				for (int x = 0; x < m_layer1.Length; x++)
				{
					if (m_layer1[x] > 0)
					{
						m_errors1[x] *= m_layer1[x] * (1 - m_layer1[x]);
					}
					else
					{
						m_errors1[x] *= - m_layer1[x] * (1 + m_layer1[x]);
					}
					// back prop the error, update the weights
					for (int y = 0; y < m_input.Length; y++)
					{
						//dpErrorsIn[y] += dpErrors1[x] * dpWt1ToIn[iWtPos];
						// update the delta and weight
						m_wt1ToIn[iWtPos] += (m_delta1ToIn[iWtPos] = m_learnRate * m_errors1[x] * m_input[y] + m_momentum * m_delta1ToIn[iWtPos]);
						//
						// need to check for exploding weights here
						//
						if (m_wt1ToIn[iWtPos] > 100.0 || m_wt1ToIn[iWtPos] < -100.0)
						{
							throw new NeuralNetworkException("Exploding weights");
						}
						if (m_wt1ToIn[iWtPos] == 0.0)
						{
							throw new NeuralNetworkException("Weight underflow");
						}
						iWtPos++;
					}
				}
				if (m_useAdaptiveLR)
				{
					delay++;
					if (dErrTotal < m_previousError && delay > 10)
					{
						m_learnRate += .00001;
						delay = 0;
					}
					else if (m_learnRate > .02)
					{
						m_learnRate -= .005 * m_learnRate;
					}
					m_previousError = dErrTotal;
				}
				if ( dErrTotal.Equals(Double.NaN) )
				{
					throw new NeuralNetworkException("Total error NaN: error=" + dErrTotal);
				}
			}
			while (dErrTotal > dErrTarget && maxCycles-- > 0);

			return dErrTotal;
		}

		public double[] Output
		{
			get
			{
				return m_output;
			}
		}

		/// <summary>  
		/// Randomize all the weight values
		/// </summary>
		private void RandomizeWeights()
		{
			// initialize the weights
			for (int x = 0; x < m_wt1ToIn.Length; x++)
			{
				m_wt1ToIn[x] = (m_rand.NextDouble() * m_wtRange) - m_wtRange;
			}
			for (int x = 0; x < m_wt2To1.Length; x++)
			{
				m_wt2To1[x] = (m_rand.NextDouble() * m_wtRange) - m_wtRange;
			}
			for (int x = 0; x < m_wtOutTo2.Length; x++)
			{
				m_wtOutTo2[x] = (m_rand.NextDouble() * m_wtRange) - m_wtRange;
			}
		}

		private void SetArray(double[] a, double v)
		{
			for (int x = 0; x < a.Length; x++)
			{
				a[x] = v;
			}
		}

		public int NetworkId
		{
			get
			{
				if ( m_networkId < 0 )
				{
					throw new NeuralNetworkException( "Network not saved" );
				}
				return m_networkId;
			}
		}

		public int InputCount
		{
			get
			{
				return m_input.Length;
			}
		}

		public override int GetHashCode()
		{
			if ( m_networkId != -1 )
			{
				return m_networkId;
			}
			return (m_learnRate.ToString() + m_momentum.ToString() + m_wtRange.ToString() + m_useAdaptiveLR.ToString()
				+ m_useAdaptiveMom.ToString()).GetHashCode() ^
				m_input.GetHashCode() ^ 
				m_layer1.GetHashCode() ^
				m_layer2.GetHashCode() ^
				m_output.GetHashCode() ^			
				m_wt1ToIn.GetHashCode() ^ 
				m_delta1ToIn.GetHashCode() ^
				m_wt2To1.GetHashCode() ^
				m_delta2To1.GetHashCode() ^ 			
				m_wtOutTo2.GetHashCode() ^
				m_deltaOutTo2.GetHashCode(); 		
		}

		public static bool operator ==(NeuralNetwork n1, NeuralNetwork n2)
		{
			if (((object)n1) == null)
			{
				return ((object)n2) == null;
			}
			return n1.Equals( n2 );
		}

		public static bool operator !=(NeuralNetwork n1, NeuralNetwork n2)
		{
			if (((object)n1) == null)
			{
				return ((object)n2) != null;
			}
			return !n1.Equals(n2);
		}

		public override bool Equals( object on2 )
		{
			NeuralNetwork n2 = on2 as NeuralNetwork;
			if ( null == (object)n2 )
			{
				return false;
			}
			if ( ! ( n2.m_input.Length == m_input.Length &&
				n2.m_layer1.Length == m_layer1.Length &&
				n2.m_layer2.Length == m_layer2.Length &&
				n2.m_learnRate == m_learnRate &&
				n2.m_momentum == m_momentum &&
				n2.m_output.Length == m_output.Length &&
				n2.m_previousError == m_previousError &&
				n2.m_useAdaptiveLR == m_useAdaptiveLR &&
				n2.m_useAdaptiveMom == m_useAdaptiveMom &&
				n2.m_wtRange == m_wtRange) )
			{
				return false;
			}
			if ( m_layer1.Length != n2.m_layer1.Length )
			{
				return false;
			}
			for ( int x = 0; x < m_layer1.Length; x++ )
			{
				if ( m_layer1[x] != n2.m_layer1[x] )
				{
					return false;
				}
			}
			if ( m_layer2.Length != n2.m_layer2.Length )
			{
				return false;
			}
			for ( int x = 0; x < m_layer2.Length; x++ )
			{
				if ( m_layer2[x] != n2.m_layer2[x] )
				{
					return false;
				}
			}
			if ( m_wt1ToIn.Length != n2.m_wt1ToIn.Length )
			{
				return false;
			}
			for ( int x = 0; x < m_wt1ToIn.Length; x++ )
			{
				if ( m_wt1ToIn[x] != n2.m_wt1ToIn[x] )
				{
					return false;
				}
			}
			if ( m_wt2To1.Length != n2.m_wt2To1.Length )
			{
				return false;
			}
			for ( int x = 0; x < m_wt2To1.Length; x++ )
			{
				if ( m_wt2To1[x] != n2.m_wt2To1[x] )
				{
					return false;
				}
			}
			if ( m_wtOutTo2.Length != n2.m_wtOutTo2.Length )
			{
				return false;
			}
			for ( int x = 0; x < m_wtOutTo2.Length; x++ )
			{
				if ( m_wtOutTo2[x] != n2.m_wtOutTo2[x] )
				{
					return false;
				}
			}
			if ( m_delta1ToIn.Length != n2.m_delta1ToIn.Length )
			{
				return false;
			}
			for ( int x = 0; x < m_delta1ToIn.Length; x++ )
			{
				if ( m_delta1ToIn[x] != n2.m_delta1ToIn[x] )
				{
					return false;
				}
			}
			if ( m_delta2To1.Length != n2.m_delta2To1.Length )
			{
				return false;
			}
			for ( int x = 0; x < m_delta2To1.Length; x++ )
			{
				if ( m_delta2To1[x] != n2.m_delta2To1[x] )
				{
					return false;
				}
			}
			if ( m_deltaOutTo2.Length != n2.m_deltaOutTo2.Length )
			{
				return false;
			}
			for ( int x = 0; x < m_deltaOutTo2.Length; x++ )
			{
				if ( m_deltaOutTo2[x] != n2.m_deltaOutTo2[x] )
				{
					return false;
				}
			}
			return true;
		}

		public int Layer1Count
		{
			get
			{
				return m_layer1.Length;
			}
		}

		public int Layer2Count
		{
			get
			{
				return m_layer2.Length;
			}
		}

		public void ResetID()
		{
			m_networkId = -1;
		}

		public static Random Rand
		{
			get
			{
				return m_rand;
			}
		}

		public DataStats[] Sensitivity(double lowInput, double hiInput)
		{
			const int steps = 100;
			double[] output = new double[m_output.Length];
			DataStats[] s = new DataStats[m_input.Length];
			double[] data = new double[steps];
			SetArray(m_input, lowInput);
			Activate(m_input, output);
			double baseline = SumArray(output);
			double step = (hiInput - lowInput) / steps;

			for (int x = 0; x < s.Length; x++)
			{
				SetArray(m_input, lowInput);
				for (int y = 0; y < 100; y++)
				{
					Activate(m_input, output);
					data[y] = Math.Abs(SumArray(output) - baseline);
					m_input[x] += step;
				}
				s[x] = new DataStats(data);
			}
			return s;
		}

		static double SumArray(double[] ar)
		{
			double sum = 0;
			for (int x = 0; x < ar.Length; x++)
			{
				sum += ar[x];
			}
			return sum;
		}
	}
}
