/*
 *   This file is part of the Maximum library.
 *
 *   Maximum is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Maximum is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Maximum.  If not, see <http://www.gnu.org/licenses/>.
 */
using System;
using System.Collections.Generic;
using System.Text;

namespace Maximum
{
	public class ProgramArgumentException : Exception
	{
		public ProgramArgumentException(string msg)
			: base(msg)
		{
		}
	}

	struct ProgramArgument
	{
		public string key;
		public string val;
	}

	/// <summary>
	/// Parse commandline argument.  UN*X or windows style swtiches can be handled.
	/// </summary>
	public class ProgramArguments
	{
		private List<ProgramArgument> m_switchList = new List<ProgramArgument>();
		private Dictionary<string, ProgramArgument> m_switchIdx = new Dictionary<string, ProgramArgument>();
		private string[] m_params;

		private void Init(string[] args, string allowedSwitches)
		{
			if (args.Length == 0)
			{
				return;
			}
			List<string> parms = new List<string>();
			ProgramArgument arg;
			arg.key = null;
			arg.val = null;
			bool win32ParamStyle = args[0][0] == '/';
			int splitCharPos;
			int nameStart;

			for (int x = 0; x < args.Length; x++)
			{
				string part = UnQuote(args[x]);
				if (win32ParamStyle)
				{
					// /switch:value
					if (part[0] != '/')
					{
						parms.Add(part);
						continue;
					}
					splitCharPos = part.IndexOf(':');
					nameStart = 1;
				}
				else
				{
					if (part[0] != '-')
					{
						parms.Add(part);
						continue;
					}
					if (part[1] == '-')
					{
						nameStart = 2;
					}
					else
					{
						nameStart = 1;
					}
					splitCharPos = part.IndexOf('=');
				}
				if (-1 == splitCharPos)
				{
					// bare switch
					arg.key = part.Substring(nameStart);
					arg.val = "";
				}
				else
				{
					arg.key = part.Substring(nameStart, splitCharPos - nameStart);
					arg.val = UnQuote(part.Substring(splitCharPos + 1));
				}
				if (null != allowedSwitches && allowedSwitches.IndexOf(arg.key) < 0)
				{
					throw new ProgramArgumentException("Unknown switch " + arg.key);
				}
				m_switchList.Add(arg);
				m_switchIdx.Add(arg.key, arg);
			}
			m_params = parms.ToArray();
		}

		public ProgramArguments(string[] args, string allowedSwitches)
		{
			Init(args, allowedSwitches);
		}

		public ProgramArguments(string[] args)
		{
			Init(args, null);
		}

		private string UnQuote(string str)
		{
			if (str[0] != '"')
			{
				return str;
			}
			if (!str.EndsWith("\""))
			{
				return str;
			}
			return str.Substring(1, str.Length - 2);
		}

		public int Count
		{
			get { return m_switchList.Count; }
		}

		public string this[int idx]
		{
			get { return m_switchList[idx].val; }
		}

		public string this[string key]
		{
			get 
			{
				if (m_switchIdx.ContainsKey(key))
				{
					return m_switchIdx[key].val;
				}
				return null;
			}
		}

		public bool HasSwitch(string key)
		{
			return m_switchIdx.ContainsKey(key);
		}

		public bool HasAnyOfSwitches(string keys, char delim)
		{
			string[] sws = keys.Split(new char[] { delim });
			for (int x = 0; x < sws.Length; x++)
			{
				if ( HasSwitch(sws[x]) )
				{
					return true;
				}
			}
			return false;
		}

		public string Switch(int idx)
		{
			return m_switchList[idx].key;
		}

		public string[] Parameters
		{
			get { return m_params; }
		}
	}
}
