/*
 *   This file is part of the Maximum library.
 *
 *   Maximum is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Maximum is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Maximum.  If not, see <http://www.gnu.org/licenses/>.
 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;

namespace Maximum
{
	/// <summary>
	/// Holder for those commonly needed string utilities.
	/// </summary>
	public class StringHelper
	{
		private StringHelper()
		{
		}

		public static bool IsHexNum( string s )
		{
			if ( 1 >= s.Length )
			{
				return Char.IsNumber(s[0]);
			}
			int pos = 0;
			if ( s[0] == '#' )
			{
				pos = 1;
			}
			else if ( s[0] == '0' && (s[1] == 'x' || s[1] == 'X') )
			{
				pos = 2;
			}
			for ( ; pos < s.Length; pos++ )
			{
				if ( Char.IsNumber(s[pos]) )
				{
					continue;
				}
				if ( (s[pos] >= 'a' && s[pos] <= 'f') || (s[pos] >= 'A' && s[pos] <= 'F') )
				{
					continue;
				}
				return false;
			}
			return true;
		}

		public static int CountOccurancesOf(string str, char ch)
		{
			int count = 0;
			for (int x = 0; x < str.Length; x++)
			{
				if (str[x] == ch)
				{
					count++;
				}
			}
			return count;
		}

		public static bool IsUpper(string str)
		{
			for (int x = 0; x < str.Length; x++)
			{
				if (!Char.IsUpper(str[x]))
				{
					return false;
				}
			}
			return true;
		}

		public static bool IsNumeric(string str)
		{
			if (str.Length == 0)
			{
				return false;
			}
			int dotcount = 0;
			for (int x = 0; x < str.Length; x++)
			{
				char ch = str[x];
				if ( ch == '.' && dotcount == 0 )
				{
					dotcount++;
					continue;
				}
				if (!Char.IsDigit(ch))
				{
					return false;
				}
			}
			return true;
		}

		public static bool IsInt(string str)
		{
			if (str.Length == 0)
			{
				return false;
			}
			for (int x = 0; x < str.Length; x++)
			{
				if (!Char.IsDigit(str[x]))
				{
					return false;
				}
			}
			return true;
		}

		public static string StripQuotes(string str)
		{
			if (str[0] == '"' && str[str.Length-1] == '"')
			{
				return str.Substring(1, str.Length - 2);
			}
			return str;
		}

		public static bool HasPath(string filename)
		{
			return filename.IndexOf('/') > -1 || filename.IndexOf("..") > -1;
		}

		public static int ParseRightInt32(string str)
		{
			StringBuilder buf = new StringBuilder(str.Length);
			for (int x = str.Length - 1; x >= 0; x--)
			{
				if (!Char.IsDigit(str[x]))
				{
					break;
				}
				buf.Insert(0, str[x]);
			}
			return Int32.Parse(buf.ToString());
		}

		public static string EnsureTrailingChar(string str, char ch)
		{
			if (str[str.Length - 1] != ch)
			{
				return str + ch.ToString();
			}
			return str;
		}

		public static string RightStr(string str, int count)
		{
			return str.Substring(str.Length - count);
		}

		public static string MidStr(string str, int start, int stop)
		{
			return str.Substring(start, stop - start);
		}

		public static decimal ParseMoney(string str)
		{
			return (str.Length == 0) ? 0 : Decimal.Parse(str.Replace("$", "").Replace(",", ""));
		}

		public static string RemoveNonNumerics(string str)
		{
			Regex nonNumerics = new Regex("[^0-9]");
			return nonNumerics.Replace(str, "");
		}

		private static char[] m_xmlSpChars = new char[] { '&', '<', '>', '"', '\'' };

		public static bool RequiresXmlEncoding(string str)
		{
			return str.IndexOfAny(m_xmlSpChars) > -1;
		}

		public static string XmlEncode(string str)
		{
			if (!RequiresXmlEncoding(str))
			{
				return str;
			}
			StringBuilder buf = new StringBuilder(str.Length + 30);
			for (int x = 0; x < str.Length; x++)
			{
				switch (str[x])
				{
					case '&':
						buf.Append("&amp;");
						break;
					case '<':
						buf.Append("&lt;");
						break;
					case '>':
						buf.Append("&gt;");
						break;
					case '"':
						buf.Append("&quot;");
						break;
					case '\'':
						buf.Append("&apos;");
						break;
					default:
						buf.Append(str[x]);
						break;
				}
			}
			return buf.ToString();
		}

		public static int IndexOfCount(string str, char ch, int count)
		{
			int pos = -1;
			while (count-- > 0 && (pos = str.IndexOf(ch, pos + 1)) > -1)
				;
			return pos;
		}
		
		public static bool EndsWith( string str, string cp )
		{
			int pos = str.Length - cp.Length;
			int cppos = 0;
			int x;
		
			if ( pos <= 0 )
			{
				return false;
			}
			for ( x = pos; x < str.Length; x++ )
			{
				if ( str[x] != cp[cppos++] )
				{
					return false;
				}
			}
			return true;
		}

		public static bool EndsWith( StringBuilder str, string cp )
		{
			int pos = str.Length - cp.Length;
			int cppos = 0;
			int x;
		
			if ( pos < 0 )
			{
				return false;
			}
			for ( x = pos; x < str.Length; x++ )
			{
				if ( str[x] != cp[cppos++] )
				{
					return false;
				}
			}
			return true;
		}
		
		public static bool AreEqual(StringBuilder sb, string str)
		{
			if ( sb.Length != str.Length )
			{
				return false;
			}
			for ( int x = 0; x < str.Length; x++ )
			{
				if ( str[x] != sb[x] )
				{
					return false;
				}
			}
			return true;
		}
	}
}
