/*
 *   This file is part of the Maximum library.
 *
 *   Maximum is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Maximum is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Maximum.  If not, see <http://www.gnu.org/licenses/>.
 */
using System;
using System.Collections.Generic;
using System.Text;

namespace Maximum
{
	/// <summary>
	/// For small lists, or lists with lots of linear scans, Vector<> can be more efficient than List<>.
	/// </summary>
	public class Vector<T>
	{
		protected T[] m_data;
		protected int m_used;

		public Vector()
		{
			m_data = new T[10];
		}

		public Vector(int size)
		{
			m_data = new T[size];
		}

		public T Pop()
		{
			if (m_used > 0) 
			{
				return m_data[--m_used];
			}
			else
			{
				return default(T);
			}		
		}
		
		public T Head()
		{
			return m_data[0];
		}
		
		public T Tail ()
		{
			return m_data[m_used - 1];
		}
		
		public int Capacity()
		{
			return m_data.Length;
		}
		
		public int AddElement(T o) 
		{
			Extend();
			m_data[m_used] = o;
			return m_used++;
		}

		public int Add(T o)
		{
			Extend();
			m_data[m_used] = o;
			return m_used++;
		}

		public void Extend() 
		{
			if (m_used >= m_data.Length) 
			{
				int space = m_data.Length;
				while (m_used >= space) 
				{
					space <<= 1;
				}
				T[] array2 = new T[space];
				int i;
				for (i = 0; (i < m_used); i++) 
				{
					array2[i] = m_data[i];
				}
				m_data = array2;
			}
		}
		
		public int Size
		{
			get { return m_used; }
		}

		public int Count
		{
			get { return m_used; }
		}
		
		public void SetSize(int s) 
		{
			if (m_data.Length < s) 
			{
				m_data = new T[s];
			}
			m_used = s;
		}
		
		public void Clear()
		{
			m_used = 0;
		}
		
		public T ElementAt(int at) 
		{
			return m_data[at];
		}

		public T this[int idx]
		{
			get { return ElementAt(idx); }
			set { SetElementAt(idx, value); }
		}

		public void SetElementAt(int at, T o) 
		{
			m_data[at] = o;
		}

		public T FirstElement() 
		{
			if (m_used > 0) 
			{
				return m_data[0];
			} else 
			{
				return default(T);
			}
		}
		
		public T LastElement()
		{
			if (m_used > 0) 
			{
				return m_data[m_used - 1];
			}
			else
			{
				return default(T);
			}
		}
		
		public void RemoveElement() 
		{
			RemoveElementAt(0);
		}
		
		public void RemoveElement(T o) 
		{
			int i;
			for (i = 0; (i < m_used); i++) 
			{
				if (m_data[i].Equals(o)) 
				{
					RemoveElementAt(i);
					return;
				}
			}
		}
		
		public int IndexOf(T o) 
		{
			int i;
			for (i = 0; (i < m_used); i++) 
			{
				if (m_data[i].Equals(o)) 
				{
					return i;
				}
			}
			return -1;
		}

		public void RemoveElementAt(int at) 
		{
			int i;
			for (i = at; (i < (m_used - 1)); i++) 
			{
				m_data[i] = m_data[i + 1];
			}
			m_used--;
		}
		
		public void InsertElementAt(T o, int at) 
		{
			int i;
			Extend();
			for (i = m_used; (i > at); i--) 
			{
				m_data[i] = m_data[i - 1];
			}
			m_data[at] = o;
			m_used++;
		}
		
		public void Copy (Vector<T> v)
		{
			Clear();
			for (int i = 0; i < v.m_used; i++)
			{
				AddElement(v.ElementAt(i));
			}
		}
		
		public void InsertBefore(T oNew, T oExisting)
		{
			int at = IndexOf(oExisting);
			if ( at < 0 )
			{
				throw new Exception("insertBefore can't find before");
				//addElement(oNew);
				//return;
			}
			InsertElementAt(oNew, at);
		}
		
		public bool IsEmpty()
		{
			return m_used == 0;
		}
	}
}
