/*
 *   This file is part of the Maximum library.
 *
 *   Maximum is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   Maximum is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Maximum.  If not, see <http://www.gnu.org/licenses/>.
 */
using System;
using System.Diagnostics;

namespace Maximum
{
	/// <summary>
	/// A ZIP + 4
	/// </summary>
	public class ZIP
	{
		private int m_zip;
		private int m_plus4;

		public ZIP(int zip)
		{
			if ( zip > 99999 )
			{
				m_zip = zip / 10000;
				m_plus4 = zip - m_zip * 10000;
				if ( m_zip > 99999 )
				{
					throw new ArgumentException( "Invalid zip code " + zip );
				}
				if ( m_plus4 > 9999 )
				{
					throw new ArgumentException( "Invalid zip code " + zip );
				}
			}
			else
			{
				m_zip = zip;
				m_plus4 = 0;
			}
		}

		public ZIP ( int zip, int plus4 )
		{
			m_zip = zip;
			m_plus4 = plus4;
			if ( m_zip > 99999 )
			{
				throw new ArgumentException( "Invalid zip code " + zip );
			}
			if ( m_plus4 > 9999 )
			{
				throw new ArgumentException( "Invalid zip code " + zip );
			}
		}

		public int Zip5
		{
			get { return m_zip; }
		}

		public int Plus4
		{
			get { return m_plus4; }
		}

		public bool HasPlus4
		{
			get { return 0 != m_plus4; }
		}

		public int ToInt()
		{
			return m_zip * 10000 + m_plus4;
		}

		public override string ToString()
		{
			if (0 != m_plus4)
			{
				return m_zip.ToString("00000") + "-" + m_plus4.ToString("0000");
			}
			return m_zip.ToString("00000");
		}

		public string ToShortString()
		{
			if (0 != m_plus4)
			{
				return m_zip.ToString("00000") + m_plus4.ToString("0000");
			}
			return m_zip.ToString("00000");
		}

		public static ZIP Parse(string zip5, string zipPlus4)
		{
			if (zip5.Length > 5 || zipPlus4.Length > 4)
			{
				throw new ArgumentException("Invalid ZIP code " + zip5 + "-" + zipPlus4);
			}
			if (!StringHelper.IsNumeric(zip5) || !StringHelper.IsNumeric(zipPlus4))
			{
				throw new ArgumentException("Invalid ZIP code " + zip5 + "-" + zipPlus4);
			}
			return new ZIP(Int32.Parse(zip5), Int32.Parse(zipPlus4));
		}

		public static ZIP Parse( string zip )
		{
			int idx;
			zip = zip.Trim();
			if ((idx = zip.IndexOf('-')) > -1)
			{
				return Parse(zip.Substring(0, idx), zip.Substring(idx + 1));
			}
			else if (zip.Length == 9)
			{
				return Parse(zip.Substring(0, 5), zip.Substring(5, 4));
			}
			if (zip.Length == 0)
			{
				return new ZIP(0);
			}
			if ( zip.Length != 5 )
			{
				throw new ArgumentException( "Invalid ZIP code " + zip );
			}
			return new ZIP( Int32.Parse( zip ) );
		}

		public static bool IsZip(string zip, string plus4)
		{
			if (!IsZip(zip))
			{
				return false;
			}
			return StringHelper.IsNumeric(zip) && zip.Length < 6 && plus4 != null && plus4.Length < 5 && StringHelper.IsNumeric(plus4);
		}

		private static char[] m_dashArray = new char[] { '-' };

		public static bool IsZip(string zip)
		{
			int dashpos = zip.IndexOf('-');
			if (dashpos > -1)
			{
				string[] parts = zip.Split(m_dashArray);
				if (parts.Length != 2)
				{
					return false;
				}
				return IsZip(parts[0], parts[1]);
			}
			if (zip.Length > 5)
			{
				return IsZip(zip.Substring(0, 5), zip.Substring(5));
			}
			return zip.Length < 6 && StringHelper.IsNumeric(zip);
		}
	}
}
